#ifndef _EPSDRAWING_H_
#define _EPSDRAWING_H_

#include "drawing.h"

class epsdrawer: public drawer {
private:
  ostream &out;
  int left,right,bottom,top; // Bounding box
  point intrinsic(point p) const {
    return point(double(left)+(right-left)*(p.x-lower_left.x)/
		 (upper_right.x-lower_left.x), 
		 double(bottom)+(top-bottom)*(p.y-lower_left.y)/
		 (upper_right.y-lower_left.y));};
  void write(point p,ostream &out) const {
    point intr=intrinsic(p);
    out<<intr.x<<" "<<intr.y;
  };
public:
  epsdrawer(ostream &o, int l,int r, int b, int t):
    drawer(point(l,b),point(r,t)),
    out(o), left(l), right(r), bottom(b), top(t) {
    out<<"%!PS-Adobe-2.0\n";
    out<<"%%BoundingBox: "<<left<<" "<<bottom<<" "<<right<<" "<<top<<"\n";
    out<<"%%EndComments\n\n";
  };
  ~epsdrawer() {out<<"showpage\n";};

  virtual void graph(const map<double,double> &g) {
    out<<"newpath\n";
    for (map<double,double>::const_iterator i=g.begin();i!=g.end();++i) {
      write(point(i->first,i->second),out);
      out<<" ";
      if (i==g.begin()) out<<"moveto\n";
      else out<<"lineto\n";
    }
    out<<"stroke\n";
    //    cerr<<"["<<g.size()<<" points graph drawn]\n";
  };

  virtual void curve(const vector<point> &g) {
    out<<"newpath\n";
    for (vector<point>::const_iterator i=g.begin();i!=g.end();++i) {
      write(*i,out);
      out<<" ";
      if (i==g.begin()) out<<"moveto\n";
      else out<<"lineto\n";
    }
    out<<"stroke\n";
    //    cerr<<"["<<g.size()<<" points curve drawn]\n";
  };

  virtual void line(point from, point to) {
    out<<"newpath ";
    write(from,out);
    out<<" moveto ";
    write(to,out);
    out<<" lineto stroke\n";
  };
  
  virtual void setlinewidth(double w) {
    out<<w<<" setlinewidth\n";
  }
};

#endif
