#define _GNU_SOURCE
#include <math.h>
#include <set>

#include "eqdraw.h"

bool operator<(point p,point q) {
  return (p.x<q.x || (p.x == q.x && p.y<q.y));
  }

double sqr(double x) {return x*x;}

double calculation_step=0.001; //relative to width
double visualization_step=0.01;
double spatial_grid=0.08;
double clipping_distance=0.1; //relative to spatial_grid

map<double,double>
DiffSolve(const function<point,double> &f,point start, drawer &draw) {
  map<double,double> graph;
  double eps=calculation_step*draw.width();
  double eps2=visualization_step*draw.width();
  graph[start.x]=start.y;
  for (int i=0;i<2;++i, eps=-eps) {
    point last=start;
    double xdrawn=start.x;
    double ydrawn=start.y;
    for (;last.x<=draw.x1() && last.x>=draw.x0();) {
      double m=f(last);
      if (fabs(m)>1000.0 || (last.y>draw.y1() || last.y<draw.y0())) {
	graph[last.x]= m*eps>0? draw.y1() : draw.y0(); 
	break;
      }
      double inc=eps*cos(atan(m));
      point curr(last.x+inc,last.y+inc*m);
      //      cerr<<x+inc<<" <- "<<graph[x+inc]<<"\n";
      if (sqrt(sqr(curr.x -xdrawn)+sqr(curr.y-ydrawn)) >= eps2) {
	graph[curr.x]=curr.y;
	xdrawn=curr.x;
	ydrawn=curr.y;
      }
      last=curr;
    }
    graph[last.x]=last.y;
  }
  //  cerr<<"graph with "<<graph.size()<<" points drawn\n";
  return graph;
};

double dist(point p, const set<point> &drawn,double min=10000.0) {
  double sq=10000.0;
  set<point>::const_iterator end=drawn.upper_bound(point(p.x+min,p.y));
  for (set<point>::const_iterator i=drawn.lower_bound(point(p.x-min,p.y));
       i!=end;++i) {
    double s=sqr(p.x - i->x)+sqr(p.y - i->y);
    if (s<sq) sq=s;
  }
  return sqrt(sq);
}

int DiffDrawAll(const function<point,double> &f, drawer &draw) {
  set<point> drawn;
  point p;
  int r=0;
  double grid=spatial_grid*draw.width();
  double d=clipping_distance*grid;
  
  for (p.x=draw.x0()+grid/2;p.x<draw.x1()-grid/2;p.x+=grid)
    for (p.y=draw.y0()+grid/2;p.y<draw.y1()-grid/2;p.y+=grid){
      if (dist(p,drawn,grid/2.0)>=grid/2.0) {
	map<double,double> graph=DiffSolve(f,p,draw);
	map<double,double>::iterator i;
	/*
	for (i=graph.lower_bound(p.x);i!=graph.end() &&
	       dist(point(i->first,i->second),drawn,grid/20)>d;++i);
	graph.erase(i,graph.end());
	for (i=graph.lower_bound(p.x);i!=graph.begin() &&
	       dist(point(i->first,i->second),drawn,grid/20)>d;--i);
	graph.erase(graph.begin(),i);
	*/
	draw.graph(graph);
	r+=graph.size();
	//	cerr<<"drawn! ("<<graph.size()<<")\n";
	for (i=graph.begin();i!=graph.end();++i)
	  drawn.insert(point(i->first,i->second));
      }
  }
  return r;
};

int FnDraw(const function<double,double> &f,drawer &draw) {
  map<double,double> graph;
  int nout=0;
  int r=0;
  double eps=calculation_step*draw.width();
  for (double x=draw.x0(); x<draw.x1();x+=eps) {
    double y=f(x);
    if (!isfinite(y) || y>draw.y1() || y<draw.y0()) nout++;
    else nout=0;
    if (nout>=2) {
      if (graph.size()>1) {
	draw.graph(graph);
	r+=graph.size();
      }
      graph.clear();
    } 
    if (isfinite(y))
      graph[x]=y;
  }
  if (graph.size()>1) {
    draw.graph(graph);
    r+=graph.size();
  }
  return r;
}

int CurveDraw(const function<double,point> &f,drawer &draw, 
	       double t0, double t1) {
  vector<point> curve;
  double eps=calculation_step*(t1-t0);
  for (double t=t0; t<t1+eps/2.0; t+=eps) {
    point p=f(t);
    curve.push_back(p);
  }
  draw.curve(curve);
  return curve.size();
}

void AxesDraw(drawer &draw) {
  draw.line(point(draw.x0(),0),point(draw.x1(),0));
  draw.line(point(0,draw.y0()),point(0,draw.y1()));
}
