/*
 * eqdraw: manu-fatto (2002)
 */

#define MYNAME "eqdraw"
#ifndef VERSION
#  define VERSION "0.2.0"
#endif


#include <iostream>
#include <map>
#include <string>
#include <strstream>
#include <algorithm>
#include <fstream>
#include <memory>

#include "epsdrawing.h"
#include "pngdrawing.h"
#include "eqdraw.h"
#include "algebra.h"
#include "parser.h"
#include "options.h"
#include "cgi.h"

//using namespace Algebra;

class exprFunction2: public function<point,double> {
  Expr E;
  typedef map<string,Expr> map_t;
  map_t map;
  Expr *first;
  Expr *second;
public:
  exprFunction2(Expr e,string indep, string dep): E(e) {
    map[indep]=0.0;
    map[dep]=0.0;
    first=&(map[indep]);
    second=&(map[dep]);
  };
  virtual ~exprFunction2(){};
  double operator()(point p) const {
    *first=Expr(p.x);
    *second=Expr(p.y);
    return (E.replace(map)).eval();
  };
};

class exprFunction1: public function<double,double> {
  Expr E;
  typedef map<string,Expr> map_t;
  map_t map;
  Expr *first;
public:
  exprFunction1(Expr e,string indep): E(e) {
    map[indep]=0.0;
    first=&(map[indep]);
  };
  virtual ~exprFunction1(){};
  double operator()(double x) const {
    *first=Expr(x);
    return (E.replace(map)).eval();
  };
};

class exprFunction3: public function <double,point> {
  Expr X,Y;
  typedef map<string,Expr> map_t;
  map_t map;
  Expr *first;
public:
  exprFunction3(Expr x, Expr y,string param): X(x), Y(y) {
    map[param]=0.0;
    first=&(map[param]);
  }
  virtual ~exprFunction3(){};
  point operator()(double t) const {
    *first = Expr(t);
    return point(X.replace(map).eval(),Y.replace(map).eval());
  };
};

bool verbose=true;

string parametric[]={"t","x","y","x0","y0"};
string *cartesian=parametric+1;

//Expr F;

typedef enum {DEFAULT,EPS,PNG,HTML} output_format_enum;
int output_format=DEFAULT; 
char *output_format_str[]={"eps","png","html",0};
int output_format_vals[]={EPS,PNG,HTML};

string output_filename="";
bool cgi_mode=false;

int image_width=400, image_height=300;
double xmin=-4, xmax=4, ymin=-3, ymax=3;
double tmin=0, tmax=1;
bool keep_ratio=true;

int eps_page_width=int(sqrt(10000.0/16.0/sqrt(2.0)) /2.54*60.0);
int eps_page_height=int(eps_page_width*sqrt(2.0)); 
int eps_x_off=(eps_page_width-image_width)/2;
int eps_y_off=(eps_page_height-image_height)/2;

Options opt;

char * cgi_options[]={"e","xmin","xmax","ymin","ymax","tmin","tmax","width","height",0};

class mystream {
  ostream *out;
public:
  mystream(){out=&cout;};
  void reset(ostream *o) {out=o;};
  void away() {
    reset(&cerr);
    if (verbose && !cgi_mode)
      cerr<<"(messages moved from stdout to stderr)\n";
  }
  ostream &operator()(){return *out;};
} out;

string bpar() {
  if (output_format==HTML) return "<p>";
  else return "";
}

string epar() {
  if (output_format==HTML) return "</p>\n";
  else return "\n";
}

string br() {
  if (output_format==HTML) return "<br>\n";
  else return "\n";
}

string bstrong() {
  if (output_format==HTML) return "<strong>";
  else return "";
}

string estrong() {
  if (output_format==HTML) return "</strong>";
  else return "";
}

void write_supported_equations(ostream &out) {
  out<<cartesian[1]<<" = f("<<cartesian[0]<<")\n"
  <<cartesian[1]<<"' = f("<<cartesian[0]<<","<<cartesian[1]<<")\n"
  <<parametric[1]<<" = f("<<parametric[0]<<"), "
  <<parametric[2]<<" = g("<<parametric[0]<<")\n";
}

void usage(Options &opt) {
  cout<<bpar()<<bstrong()<<MYNAME " ver. " VERSION 
    " - manu-fatto"<<estrong()<<br();
  cout<<"usage:"<<br();
  if (output_format==HTML) cout<<"<pre>\n";
  cout<<MYNAME<<" [options]\n";
  opt.usage(cout);
  cout<<"supported equations:\n";
  write_supported_equations(cout);
  if (output_format==HTML) cout<<"</pre>";
  cout<<epar();
};

class drawer_ptr {
  drawer *draw;
  FILE *fp;
  ostream *o;
  bool prepared;
public:
  drawer_ptr(): draw(0), fp(0), o(o), prepared(false) {};
  void prepare() {
    if (!prepared) open();
    prepared=true;
  };
  void open() {
    close();
    fp=0;
    o=0;
    draw=0;
    if (cgi_mode) output_filename="";
    if (output_format==DEFAULT && output_filename !="") {
      if (verbose)
	out()<<"writing to file "<<output_filename<<"\n";
      string::size_type d=output_filename.rfind('.');
      if (d!=string::npos) {
	string ext=string(output_filename,d+1);
	if (ext=="png" || ext=="PNG") output_format=PNG;
	if (ext=="ps" || ext=="eps" || ext=="PS" || ext=="EPS")
	  output_format=EPS;
      }
    }
    if (output_format==DEFAULT) output_format=EPS;
    if (output_format==PNG) {
      FILE *f=stdout;
      if (output_filename!="") {
	fp=fopen(output_filename.c_str(),"wb");
	if (!fp) throw runtime_error("Cannot write file "+output_filename);
	f=fp;
      } else {out.away();};
      draw=new pngdrawer(f,image_width,image_height);
    } else if (output_format==EPS) {
      ostream *oo=&cout;
      if (output_filename!="") {
	o=new ofstream(output_filename.c_str());
	if (!o || !*o) 
	  throw runtime_error("Cannot write file "+output_filename);
	oo=o;
      } else {out.away();}; 
      draw=new epsdrawer(*oo,eps_x_off,eps_x_off+image_width,
			      eps_y_off,eps_y_off+image_height);
    } else if (output_format==HTML) {
      string url=Cgi::GetEnv("SCRIPT_NAME")+string("?");
      for (int i=0;cgi_options[i];++i)
	url+=opt[cgi_options[i]].url();
      cout<<"<img width="<<image_width<<" height="<<image_height
	  <<" src=\""<<url<<"t=png&\"><br>\n"
	"(<a href=\""<<url<<"t=eps&\">versione postscript</a>)\n";
    }
    if (output_format!=HTML) {
      draw->window(point(xmin,ymin),point(xmax,ymax));
      // draw axes 
      draw->setlinewidth(0.5);
      draw->line(point(draw->x0(),0),point(draw->x1(),0));
      draw->line(point(0,draw->y0()),point(0,draw->y1()));
      draw->setlinewidth(1.0);
    }
  };
  void close() {
    delete draw;
    draw=0;
    if (fp) fclose(fp);
    fp=0;
    delete o;
    o=0;
  };
  ~drawer_ptr(){close();};
  drawer* operator->() {prepare();return draw;};
  drawer& operator*() {prepare();return *draw;};
  operator bool() {prepare();return draw!=0;};
} output_draw;

Expr string2Expr(const char *s) {
  Expr F;
  istrstream in(s);
  in>>F;
  return F;
};

int order(string var,string respect) {
  if (var.size()<respect.size()) return -1;
  if (string(var,0,respect.size()) == respect) {
    string::size_type i;
    for(i=respect.size();i<var.size() && var[i]=='\'';++i);
    if (i==var.size()) return i-respect.size();
  }
  return -1;
}

int order(set<string> &vars, const string* respect, int n) {
  set<string>::const_iterator i;
  int ord=0;
  for (i=vars.begin();i!=vars.end();++i) {
    int k=-1;
    for (int j=0;j<n;++j) {
      int n=order(*i,respect[j]);
      if (n>=0) {
	if (k<0) k=n;
	else if (n<k) k=n;
      }
    }
    if (k<0) return -1;
    if (k>ord) ord=k;
  }
  return ord;
}

void draw(Expr F) {
  set<string> vars=F.var_list();
  if (output_format==HTML) output_draw.prepare();
  if (false) {
    out()<<"variabili:";
    for (set<string>::iterator i=vars.begin();i!=vars.end();++i)
      out()<<" "<<*i;
    out()<<"\n";
  }
  if (order(vars,cartesian,2)==1) { 
    //equazione differenziale del prim'ordine
    Match m(F,Equation(Expr("left"),Expr("right")));
    if (m) {
      vars=m["right"].var_list();
      if (m["left"].var_name()== cartesian[1]+"'" &&
	  order(vars,cartesian,2)==0) {
	//in forma normale y'=f(x,y)
	if (verbose) { 
	  out()<<bpar()<<
	    "Equazione differenziale del prim'ordine in forma normale: "
	       <<bstrong()<<F<<estrong()<<epar();
	}
	exprFunction2 eF(m["right"],cartesian[0],cartesian[1]);
	if (output_draw) {
	  int n=DiffDrawAll(eF,*output_draw);
	  if (verbose) 
	    out()<<n<<" lines drawn\n";
	}
	return;
      }
    }
  } else if (order(vars,cartesian,2)==0) {
    //equazione algebrica
    Match m(F,Equation(Expr("left"),Expr("right")));
    vars=m["right"].var_list();
    if (m["left"].var_name() == cartesian[1] &&
	order(vars,cartesian,1)==0) {
      //in forma normale y=f(x)
      if (verbose) 
	out()<<bpar()<<"Grafico di funzione: "<<bstrong()<<F<<estrong()<<
	  epar();
      exprFunction1 eF(m["right"],cartesian[0]);
      if (output_draw) {
	int n=FnDraw(eF,*output_draw);
	if (verbose)
	  out()<<n<<" lines drawn\n";
      }
      return;
    }
  } else if (order(vars,parametric,3)==0) {
    Match s(F,MakeSystem(Expr("first"),Expr("second")));
    if (s) { // systema
      string which="first";
      set<string> param;
      vars=param;
      Expr x,y;
      for (int i=0;i<2;++i) {
	Match m(s[which],Equation(Expr("left"),Expr("right")));
	param=m["right"].var_list();
	if (order(param,parametric,1)!=0) break;
	string left=m["left"].var_name();
	if (left == cartesian[0]) x=m["right"];
	else if (left == cartesian[1]) y=m["right"];
	vars.insert(m["left"].var_name());
	which="second";
      }
      if (order(vars,cartesian,2)==0) {
	if (verbose) 
	  out()<<bpar()<<"Supporto della curva: "<<bstrong()
	       <<F<<estrong()<<epar();
	exprFunction3 eF(x,y,parametric[0]);
	if (output_draw) {
	  int n=CurveDraw(eF,*output_draw,tmin,tmax);
	  if (verbose)
	    out()<<n<<" lines drawn\n";
	}
	return;
      }
    }
  }
  out()<<
    "Non so trattare questo tipo di equazioni.\n"
    "Eq: "<<F<<"\n";
  if (verbose) {
    out()<<"Equazioni correntemente supportate:\n";
    write_supported_equations(out());
  }
};

void Optdraw(Options &opt,string s) {
  if (s!="")
    draw(string2Expr(s.c_str()));
};

void set_options() {
  opt.add(Option(new OptNormal("o","output",
			       "-o|--output <filename>: set output file name"),
		 new OptStringRef(output_filename)));
  opt.add(Option(new OptNormal("h","help",
			       "-h|--help: writes help information"),
		 new OptFun(usage,-100)));
  opt.add(Option(new OptNormal("t","type",
				"-t|--type [eps|png|html]: output format"),
		 new OptCases(output_format_str,output_format_vals,
			      output_format)));
  opt.add(Option(new OptNormal("cgi","cgi",
				"-cgi: cgi mode on"),
		 new OptBool(cgi_mode)));
  opt.add(Option(new OptNormal("v","verbose",
				"-v|--verbose: be verbose"),
		  new OptBool(verbose)));

  opt.add(Option(new OptNormal("xmin","xmin",
			       "-xmin <val>"),
		 new OptDoubleRef(xmin)));  
  opt.add(Option(new OptNormal("xmax","xmax",
			       "-xmax <val>"),
		 new OptDoubleRef(xmax)));
  opt.add(Option(new OptNormal("ymin","ymin",
			       "-ymin <val>"),
		 new OptDoubleRef(ymin)));
  opt.add(Option(new OptNormal("ymax","ymax",
			       "-ymax <val>"),
		 new OptDoubleRef(ymax)));
  opt.add(Option(new OptNormal("tmin","tmin",
			       "-tmin <val>"),
		 new OptDoubleRef(tmin)));
  opt.add(Option(new OptNormal("tmax","tmax",
			       "-tmax <val>"),
		 new OptDoubleRef(tmax)));

  opt.add(Option(new OptNormal("width","image-width",
			       "-width|--image-width <val>"),
		 new OptIntRef(image_width)));
  opt.add(Option(new OptNormal("height","image-height",
			       "-height|--image-height <val>"),
		 new OptIntRef(image_height)));

  
  //ultima!
  opt.add(Option(new OptNormal("e","equation","-e|--equation <equation>"),
		 new OptFunForeachString(Optdraw)));

};

Cgi *cgi_ptr=0;

void html_headers() {
  out()<<"<html><head><title>Equation drawer (manu-fatto)</title></head>\n"
    "<body><h1>Equation drawer</h1>\n";
}

void html_footers() {
  if (cgi_ptr && !cgi_ptr->empty()) {
    out()<<"<hr><p><strong>add or modify equations:</strong><br>\n"
	 <<"<form method=\"POST\" action=\""<<Cgi::GetEnv("SCRIPT_NAME")<<"\">\n"
	 <<"Equations: "<<opt["e"].form()
	 <<"<input name=\"e\"><br>\n";
    for (int i=1;cgi_options[i];++i) {
      if (i%2==1) out()<<"<br>";
      out()<<cgi_options[i]<<": "<<opt[cgi_options[i]].form()<<"\n";
    }
    //       <<"<input type=\"reset\" value=\"clear\">\n"
    out()<<"<br><input type=\"submit\" value=\"draw\"><br>\n"
	 <<"</form></p><hr>\n";
  }

  char* esempi[]={"y=x^2/3+sin(5*x)","x=t*cos(10*Pi*t), y=t*cos(6*Pi*t)","y'=(y-x)*(x^2*y+1)",0};

  out()<<"<p><strong>esempi:</strong>:"
    "<ul>";
  for (int i=0;esempi[i];++i) {
    out()<<"<li> <form method=\"POST\" action=\""
	 <<Cgi::GetEnv("SCRIPT_NAME")<<"\">\n"
	 <<"<input name='e' value=\""<<esempi[i]<<"\">"
	 <<"<input type='submit' value='prova!'></form>\n";
  }
  out()<<"</ul></p>\n";
  out()<<"<p><strong>new drawing:</strong><br>\n"
       <<"<form action=\""<<Cgi::GetEnv("SCRIPT_NAME")<<"\">\n"
       <<"Equation: "
       <<"<input name=\"e\"><br>\n"
    //       <<"<input type=\"reset\" value=\"clear\">\n"
       <<"<input type=\"submit\" value=\"draw\"><br>\n"
       <<"</form></p><hr>\n";

  usage(opt);
  out()<<"</body></html>\n"; 
}

int main(int argc, char *argv[]) {
  try {
    set_options();
    {
      cgi_mode=false;
      string myname=argv[0];
      string::size_type f=myname.rfind('.');
      if (f!=string::npos && string(myname,f) == ".cgi") cgi_mode=true;
      if (getenv("REQUEST_METHOD")!=0) cgi_mode=true;    
    }
    if (cgi_mode) { // sono un cgi!
      output_format=HTML;
      try {
	Cgi the_cgi;
	cgi_ptr=&the_cgi;
	opt.parse(the_cgi);
	if (output_format!=HTML) verbose=false;
      } catch (runtime_error e) {
	cout<<"content-type: text/html\n\n"
	  "<html><body><h1>Cgi Error!</h1>\n"
	    <<e.what()
	    <<"</html>\n";
	throw e;
      }
    } else opt.parse(argc,argv);
    
    if (cgi_mode) {
      output_filename="";
      if (output_format==HTML) {
	cout<<"content-type: text/html\n\n";
      } else if (output_format==PNG) {
	cout<<"content-type: image/png\n\n";
      } else if (output_format==EPS) {
	cout<<"content-type: application/postscript\n\n"; 
      }
    }
    if (output_format==HTML) html_headers();
    try {
      opt.exec();
    } catch (runtime_error e) {
      if (cgi_mode && output_format==HTML) 
	cout<<"<br><br><p><strong>Runtime error:</strong> "<<e.what()<<"\n";
      else throw e;
    }
    if (output_format==HTML)
      html_footers();
  } catch (runtime_error e) {
    cerr<<"Runtime error: "<<e.what()<<"\n";
  }
}
