#include <vector>
#include <set>
#include <string>
#include <algorithm>
#include "options.h"

bool OptNormal::parse(Options &opt, const CgiItem &item, OptValue &val) {
  return item.name==myname && val.parse(item.value);
};
  
bool OptNormal::parse(Options &opt, int argc,char *argv[],int &count, 
		      OptValue &val) {
  string s=argv[count];
  string::size_type f,t;
  for (f=0;f<s.size() && s[f]=='-';f++);
  for (t=f;t<s.size() && s[t]!='=';t++);
  string name=string(s,f,t-f);
  //  cout<<"name="<<name<<"; myname=="<<myname<<";\n";
  if (!(name==myname || name==myalias)) return false;
  if (t==s.size() && val.need_value() && count<argc) {
    if (val.parse(argv[count+1])) {
      count+=2; return true;
    }
  }
  if ((t==s.size() && val.parse("")) 
      || val.parse(string(s,t+1))) {
    count++;
    return true;
  }
  return false;
}

string OptNormal::usage() {
  return mydescription;
}

bool Option::parse(Options &op, int argc, char *argv[], int &count) {
  return mytype->parse(op,argc,argv,count,*myvalue);
};

bool Option::parse(Options &op, const CgiItem &item) {
  return mytype->parse(op,item,*myvalue);
};

void Options::parse(int argc, char *argv[]) {
  int count=1;
  while(count<argc) {
    list_type::iterator i;
    for (i=list.begin();i!=list.end();++i) {
      if (i->parse(*this,argc,argv,count)) break;
    }
    if (i==list.end()) {
      throw OptError(string("cannot parse cmd line option ")+argv[count]);
    }
  }
};

void Options::parse(Cgi &cgi) {
  CgiItem item;
  while(cgi.GetItem(item)) {
    list_type::iterator i;
    for (i=list.begin();i!=list.end();++i) {
      if (i->parse(*this,item)) break;
    }
    if (i==list.end()) {
      throw OptError("cannot parse cmd line option "+item.name);
    }
  }
};

void Options::exec() {
  set<int> stages;
  for(list_type::iterator i=list.begin();i!=list.end();++i) 
    stages.insert(i->execution_stage());
  for(set<int>::iterator stage=stages.begin();stage!=stages.end();++stage) {
    for (list_type::iterator i=list.begin();i!=list.end();++i) {
      if (i->execution_stage()==*stage)
	i->exec(*this);
    }
  }
}

void Options::usage(ostream &out) {
  for (list_type::iterator i=list.begin();i!=list.end();++i) {
    if (i->usage()!="")
      out<<i->usage()<<"\n";
  }
};

void OptFunForeachString::exec(Options &opt) {
  for (vector<string>::iterator i=list.begin();i!=list.end();++i) {
    (*myfn)(opt,*i);
  }
};

OptDefault::OptDefault(string name,string alias, string description):
    OptNormal(name,alias,description) {};

bool OptDefault::parse(Options &opt,int argc,char *argv[],int &count,
		       OptValue &val) {
  if (OptNormal::parse(opt,argc,argv,count,val)) return true;
  if (val.parse(argv[count])) {
    count++;
    return true;
  }
  return false;
};

Option& Options::operator[](const string &name) {
  list_type::iterator i;
  for (i=list.begin();i!=list.end() && i->name()!=name;++i)
    if (i==list.end()) throw invalid_argument("option name");
  return *i;
};

OptCases::OptCases(char *list[], int *vals, int &ref): myref(ref) {
  for (int i=0;list[i];++i)
    mymap[list[i]]=vals[i];
};

bool OptCases::parse(string s) {
  map<string,int>::const_iterator i=mymap.find(s);
  if (i!=mymap.end()) {
    myref=i->second;
    return true;
  }
  throw OptError("unexpected value "+s);
}

string OptCases::value() const {
  map<string,int>::const_iterator i;
  for (i=mymap.begin(); i!=mymap.end() && i->second!=myref;++i);
  if (i!=mymap.end()) return i->first;
  return "";
};

string OptFunForeachString::url(const OptBaseType &base) const {
  string r="";
  for (vector<string>::const_iterator i=list.begin();i!=list.end();++i)
    r+=base.url()+"="+Cgi::escape_url(*i)+"&";
  return r;
};

string OptValue::url(const OptBaseType &base) const {
  return base.url()+"="+Cgi::escape_url(value())+"&";}

string OptValue::form(const OptBaseType &base) const {
  char s[40];
  sprintf(s,"%d",value().size()+2);
  return "<input name=\""+base.name()+"\" cols=\""+string(s)
    +"\" value=\""+value()
    +"\">";
}


string OptFunForeachString::form(const OptBaseType &base) const {
  string r="";
  for (vector<string>::const_iterator i=list.begin();i!=list.end();++i)
    if (*i!="")
      r+="<!--pre--><input name=\""+base.name()+"\" value=\""+*i
	+"\"><!--post-->";
  return r;
};
